@@ -27,16 +27,14 @@ import (
 	"k8s.io/apimachinery/pkg/util/intstr"
 
 	"github.com/blang/semver"
-	monitoringv1 "github.com/coreos/prometheus-operator/pkg/apis/monitoring/v1"
-	"github.com/coreos/prometheus-operator/pkg/k8sutil"
-	"github.com/coreos/prometheus-operator/pkg/operator"
 	"github.com/pkg/errors"
+	monitoringv1 "github.com/prometheus-operator/prometheus-operator/pkg/apis/monitoring/v1"
+	"github.com/prometheus-operator/prometheus-operator/pkg/k8sutil"
+	"github.com/prometheus-operator/prometheus-operator/pkg/operator"
 )
 
 const (
 	governingServiceName            = "prometheus-operated"
-	DefaultPrometheusVersion        = "v2.7.1"
-	DefaultThanosVersion            = "v0.8.1"
 	defaultRetention                = "24h"
 	defaultReplicaExternalLabelName = "prometheus_replica"
 	storageDir                      = "/prometheus"
@@ -61,81 +59,30 @@ var (
 		managedByOperatorLabel: managedByOperatorLabelValue,
 	}
 	probeTimeoutSeconds int32 = 3
-
-	CompatibilityMatrix = []string{
-		"v1.4.0",
-		"v1.4.1",
-		"v1.5.0",
-		"v1.5.1",
-		"v1.5.2",
-		"v1.5.3",
-		"v1.6.0",
-		"v1.6.1",
-		"v1.6.2",
-		"v1.6.3",
-		"v1.7.0",
-		"v1.7.1",
-		"v1.7.2",
-		"v1.8.0",
-		"v2.0.0",
-		"v2.2.1",
-		"v2.3.1",
-		"v2.3.2",
-		"v2.4.0",
-		"v2.4.1",
-		"v2.4.2",
-		"v2.4.3",
-		"v2.5.0",
-		"v2.6.0",
-		"v2.6.1",
-		"v2.7.0",
-		"v2.7.1",
-		"v2.7.2",
-		"v2.8.1",
-		"v2.9.2",
-		"v2.10.0",
-		"v2.11.0",
-	}
 )
 
 func makeStatefulSet(
 	p monitoringv1.Prometheus,
-	config *Config,
+	config *operator.Config,
 	ruleConfigMapNames []string,
 	inputHash string,
 ) (*appsv1.StatefulSet, error) {
 	// p is passed in by value, not by reference. But p contains references like
 	// to annotation map, that do not get copied on function invocation. Ensure to
 	// prevent side effects before editing p by creating a deep copy. For more
-	// details see https://github.com/coreos/prometheus-operator/issues/1659.
+	// details see https://github.com/prometheus-operator/prometheus-operator/issues/1659.
 	p = *p.DeepCopy()
 
-	// TODO(fabxc): is this the right point to inject defaults?
-	// Ideally we would do it before storing but that's currently not possible.
-	// Potentially an update handler on first insertion.
-
-	if p.Spec.BaseImage == "" {
-		p.Spec.BaseImage = config.PrometheusDefaultBaseImage
-	}
-	if p.Spec.Version == "" {
-		p.Spec.Version = DefaultPrometheusVersion
-	}
-	if p.Spec.Thanos != nil && p.Spec.Thanos.Version == nil {
-		v := DefaultThanosVersion
-		p.Spec.Thanos.Version = &v
+	promVersion := operator.StringValOrDefault(p.Spec.Version, operator.DefaultPrometheusVersion)
+	parsedVersion, err := semver.ParseTolerant(promVersion)
+	if err != nil {
+		return nil, errors.Wrap(err, "failed to parse prometheus version")
 	}
 
 	if p.Spec.PortName == "" {
 		p.Spec.PortName = defaultPortName
 	}
 
-	versionStr := strings.TrimLeft(p.Spec.Version, "v")
-
-	version, err := semver.Parse(versionStr)
-	if err != nil {
-		return nil, errors.Wrap(err, "parse version")
-	}
-
 	if p.Spec.Replicas == nil {
 		p.Spec.Replicas = &minReplicas
 	}
@@ -152,7 +99,7 @@ func makeStatefulSet(
 	}
 	_, memoryRequestFound := p.Spec.Resources.Requests[v1.ResourceMemory]
 	memoryLimit, memoryLimitFound := p.Spec.Resources.Limits[v1.ResourceMemory]
-	if !memoryRequestFound && version.Major == 1 {
+	if !memoryRequestFound && parsedVersion.Major == 1 {
 		defaultMemoryRequest := resource.MustParse("2Gi")
 		compareResult := memoryLimit.Cmp(defaultMemoryRequest)
 		// If limit is given and smaller or equal to 2Gi, then set memory
@@ -165,7 +112,7 @@ func makeStatefulSet(
 		}
 	}
 
-	spec, err := makeStatefulSetSpec(p, config, ruleConfigMapNames)
+	spec, err := makeStatefulSetSpec(p, config, ruleConfigMapNames, parsedVersion)
 	if err != nil {
 		return nil, errors.Wrap(err, "make StatefulSet spec")
 	}
@@ -247,7 +194,7 @@ func makeStatefulSet(
 	return statefulset, nil
 }
 
-func makeEmptyConfigurationSecret(p *monitoringv1.Prometheus, config Config) (*v1.Secret, error) {
+func makeEmptyConfigurationSecret(p *monitoringv1.Prometheus, config operator.Config) (*v1.Secret, error) {
 	s := makeConfigSecret(p, config)
 
 	s.ObjectMeta.Annotations = map[string]string{
@@ -257,7 +204,7 @@ func makeEmptyConfigurationSecret(p *monitoringv1.Prometheus, config Config) (*v
 	return s, nil
 }
 
-func makeConfigSecret(p *monitoringv1.Prometheus, config Config) *v1.Secret {
+func makeConfigSecret(p *monitoringv1.Prometheus, config operator.Config) *v1.Secret {
 	boolTrue := true
 	return &v1.Secret{
 		ObjectMeta: metav1.ObjectMeta{
@@ -280,7 +227,7 @@ func makeConfigSecret(p *monitoringv1.Prometheus, config Config) *v1.Secret {
 	}
 }
 
-func makeStatefulSetService(p *monitoringv1.Prometheus, config Config) *v1.Service {
+func makeStatefulSetService(p *monitoringv1.Prometheus, config operator.Config) *v1.Service {
 	p = p.DeepCopy()
 
 	if p.Spec.PortName == "" {
@@ -328,18 +275,20 @@ func makeStatefulSetService(p *monitoringv1.Prometheus, config Config) *v1.Servi
 	return svc
 }
 
-func makeStatefulSetSpec(p monitoringv1.Prometheus, c *Config, ruleConfigMapNames []string) (*appsv1.StatefulSetSpec, error) {
+func makeStatefulSetSpec(p monitoringv1.Prometheus, c *operator.Config, ruleConfigMapNames []string,
+	version semver.Version) (*appsv1.StatefulSetSpec, error) {
 	// Prometheus may take quite long to shut down to checkpoint existing data.
 	// Allow up to 10 minutes for clean termination.
 	terminationGracePeriod := int64(600)
 
-	versionStr := strings.TrimLeft(p.Spec.Version, "v")
-
-	version, err := semver.Parse(versionStr)
+	baseImage := operator.StringValOrDefault(p.Spec.BaseImage, operator.DefaultPrometheusBaseImage)
+	if p.Spec.Image != nil && strings.TrimSpace(*p.Spec.Image) != "" {
+		baseImage = *p.Spec.Image
+	}
+	prometheusImagePath, err := operator.BuildImagePath(baseImage, p.Spec.Version, p.Spec.Tag, p.Spec.SHA)
 	if err != nil {
-		return nil, errors.Wrap(err, "parse version")
+		return nil, err
 	}
-
 	promArgs := []string{
 		"-web.console.templates=/etc/prometheus/consoles",
 		"-web.console.libraries=/etc/prometheus/console_libraries",
@@ -441,11 +390,6 @@ func makeStatefulSetSpec(p monitoringv1.Prometheus, c *Config, ruleConfigMapName
 		}
 	}
 
-	var securityContext *v1.PodSecurityContext = nil
-	if p.Spec.SecurityContext != nil {
-		securityContext = p.Spec.SecurityContext
-	}
-
 	if p.Spec.EnableAdminAPI {
 		promArgs = append(promArgs, "-web.enable-admin-api")
 	}
@@ -477,6 +421,10 @@ func makeStatefulSetSpec(p monitoringv1.Prometheus, c *Config, ruleConfigMapName
 		}
 	}
 
+	if version.GTE(semver.MustParse("2.8.0")) && p.Spec.AllowOverlappingBlocks {
+		promArgs = append(promArgs, "-storage.tsdb.allow-overlapping-blocks")
+	}
+
 	var ports []v1.ContainerPort
 	if p.Spec.ListenLocal {
 		promArgs = append(promArgs, "-web.listen-address=127.0.0.1:9090")
@@ -496,12 +444,6 @@ func makeStatefulSetSpec(p monitoringv1.Prometheus, c *Config, ruleConfigMapName
 		}
 	}
 
-	localReloadURL := &url.URL{
-		Scheme: "http",
-		Host:   c.LocalHost + ":9090",
-		Path:   path.Clean(webRoutePrefix + "/-/reload"),
-	}
-
 	volumes := []v1.Volume{
 		{
 			Name: "config",
@@ -565,6 +507,7 @@ func makeStatefulSetSpec(p monitoringv1.Prometheus, c *Config, ruleConfigMapName
 		},
 	}
 
+	promVolumeMounts = append(promVolumeMounts, p.Spec.VolumeMounts...)
 	for _, name := range ruleConfigMapNames {
 		promVolumeMounts = append(promVolumeMounts, v1.VolumeMount{
 			Name:      name,
@@ -606,24 +549,6 @@ func makeStatefulSetSpec(p monitoringv1.Prometheus, c *Config, ruleConfigMapName
 		})
 	}
 
-	configReloadVolumeMounts := []v1.VolumeMount{
-		{
-			Name:      "config",
-			MountPath: confDir,
-		},
-		{
-			Name:      "config-out",
-			MountPath: confOutDir,
-		},
-	}
-
-	configReloadArgs := []string{
-		fmt.Sprintf("--log-format=%s", c.LogFormat),
-		fmt.Sprintf("--reload-url=%s", localReloadURL),
-		fmt.Sprintf("--config-file=%s", path.Join(confDir, configFilename)),
-		fmt.Sprintf("--config-envsubst-file=%s", path.Join(confOutDir, configEnvsubstFilename)),
-	}
-
 	const localProbe = `if [ -x "$(command -v curl)" ]; then curl %s; elif [ -x "$(command -v wget)" ]; then wget -q -O /dev/null %s; else exit 1; fi`
 
 	var livenessProbeHandler v1.Handler
@@ -698,6 +623,10 @@ func makeStatefulSetSpec(p monitoringv1.Prometheus, c *Config, ruleConfigMapName
 
 	podAnnotations := map[string]string{}
 	podLabels := map[string]string{}
+	podSelectorLabels := map[string]string{
+		"app":        "prometheus",
+		"prometheus": p.Name,
+	}
 	if p.Spec.PodMetadata != nil {
 		if p.Spec.PodMetadata.Labels != nil {
 			for k, v := range p.Spec.PodMetadata.Labels {
@@ -711,80 +640,60 @@ func makeStatefulSetSpec(p monitoringv1.Prometheus, c *Config, ruleConfigMapName
 		}
 	}
 
-	podLabels["app"] = "prometheus"
-	podLabels["prometheus"] = p.Name
+	for k, v := range podSelectorLabels {
+		podLabels[k] = v
+	}
 
+	finalSelectorLabels := c.Labels.Merge(podSelectorLabels)
 	finalLabels := c.Labels.Merge(podLabels)
 
 	var additionalContainers []v1.Container
 
-	if len(ruleConfigMapNames) != 0 {
-		container := v1.Container{
-			Name:  "rules-configmap-reloader",
-			Image: c.ConfigReloaderImage,
-			Args: []string{
-				fmt.Sprintf("--webhook-url=%s", localReloadURL),
-			},
-			VolumeMounts:             []v1.VolumeMount{},
-			Resources:                v1.ResourceRequirements{Limits: v1.ResourceList{}},
-			TerminationMessagePolicy: v1.TerminationMessageFallbackToLogsOnError,
-		}
-
-		if c.ConfigReloaderCPU != "0" {
-			container.Resources.Limits[v1.ResourceCPU] = resource.MustParse(c.ConfigReloaderCPU)
-		}
-		if c.ConfigReloaderMemory != "0" {
-			container.Resources.Limits[v1.ResourceMemory] = resource.MustParse(c.ConfigReloaderMemory)
-		}
-
-		for _, name := range ruleConfigMapNames {
-			mountPath := rulesDir + "/" + name
-			container.VolumeMounts = append(container.VolumeMounts, v1.VolumeMount{
-				Name:      name,
-				MountPath: mountPath,
-			})
-			container.Args = append(container.Args, fmt.Sprintf("--volume-dir=%s", mountPath))
-		}
-
-		additionalContainers = append(additionalContainers, container)
-	}
-
 	disableCompaction := p.Spec.DisableCompaction
 	if p.Spec.Thanos != nil {
-		// Version is used by default.
-		// If the tag is specified, we use the tag to identify the container image.
-		// If the sha is specified, we use the sha to identify the container image,
-		// as it has even stronger immutable guarantees to identify the image.
-		thanosBaseImage := c.ThanosDefaultBaseImage
-		if p.Spec.Thanos.BaseImage != nil {
-			thanosBaseImage = *p.Spec.Thanos.BaseImage
+		thBaseImage := operator.StringPtrValOrDefault(p.Spec.Thanos.BaseImage, operator.DefaultThanosBaseImage)
+		thVersion := operator.StringPtrValOrDefault(p.Spec.Thanos.Version, operator.DefaultThanosVersion)
+		thTag := operator.StringPtrValOrDefault(p.Spec.Thanos.Tag, "")
+		thSHA := operator.StringPtrValOrDefault(p.Spec.Thanos.SHA, "")
+		thanosImage, err := operator.BuildImagePath(thBaseImage, thVersion, thTag, thSHA)
+		if err != nil {
+			return nil, errors.Wrap(err, "failed to build image path")
 		}
-		thanosImage := fmt.Sprintf("%s:%s", thanosBaseImage, *p.Spec.Thanos.Version)
-		if p.Spec.Thanos.Tag != nil {
-			thanosImage = fmt.Sprintf("%s:%s", thanosBaseImage, *p.Spec.Thanos.Tag)
-		}
-		if p.Spec.Thanos.SHA != nil {
-			thanosImage = fmt.Sprintf("%s@sha256:%s", thanosBaseImage, *p.Spec.Thanos.SHA)
-		}
-		if p.Spec.Thanos.Image != nil && *p.Spec.Thanos.Image != "" {
+		// If the image path is set in the custom resource, override other image settings.
+		if p.Spec.Thanos.Image != nil && strings.TrimSpace(*p.Spec.Thanos.Image) != "" {
 			thanosImage = *p.Spec.Thanos.Image
 		}
+
 		bindAddress := "[$(POD_IP)]"
 		if p.Spec.Thanos.ListenLocal {
 			bindAddress = "127.0.0.1"
 		}
 
+		thanosArgs := []string{"sidecar",
+			fmt.Sprintf("--prometheus.url=http://%s:9090%s", c.LocalHost, path.Clean(webRoutePrefix)),
+			fmt.Sprintf("--tsdb.path=%s", storageDir),
+			fmt.Sprintf("--grpc-address=%s:10901", bindAddress),
+			fmt.Sprintf("--http-address=%s:10902", bindAddress),
+		}
+
+		if p.Spec.Thanos.GRPCServerTLSConfig != nil {
+			tls := p.Spec.Thanos.GRPCServerTLSConfig
+			if tls.CertFile != "" {
+				thanosArgs = append(thanosArgs, "--grpc-server-tls-cert="+tls.CertFile)
+			}
+			if tls.KeyFile != "" {
+				thanosArgs = append(thanosArgs, "--grpc-server-tls-key="+tls.KeyFile)
+			}
+			if tls.CAFile != "" {
+				thanosArgs = append(thanosArgs, "--grpc-server-tls-client-ca="+tls.CAFile)
+			}
+		}
+
 		container := v1.Container{
 			Name:                     "thanos-sidecar",
 			Image:                    thanosImage,
 			TerminationMessagePolicy: v1.TerminationMessageFallbackToLogsOnError,
-			Args: []string{
-				"sidecar",
-				fmt.Sprintf("--prometheus.url=http://%s:9090%s", c.LocalHost, path.Clean(webRoutePrefix)),
-				fmt.Sprintf("--tsdb.path=%s", storageDir),
-				fmt.Sprintf("--grpc-address=%s:10901", bindAddress),
-				fmt.Sprintf("--http-address=%s:10902", bindAddress),
-			},
+			Args:                     thanosArgs,
 			Env: []v1.EnvVar{
 				{
 					Name: "POD_IP",
@@ -828,11 +737,29 @@ func makeStatefulSetSpec(p monitoringv1.Prometheus, c *Config, ruleConfigMapName
 			disableCompaction = true
 		}
 
-		if p.Spec.LogLevel != "" {
-			container.Args = append(container.Args, fmt.Sprintf("--log.level=%s", p.Spec.LogLevel))
+		if p.Spec.Thanos.TracingConfig != nil {
+			container.Args = append(container.Args, "--tracing.config=$(TRACING_CONFIG)")
+			container.Env = append(container.Env, v1.EnvVar{
+				Name: "TRACING_CONFIG",
+				ValueFrom: &v1.EnvVarSource{
+					SecretKeyRef: p.Spec.Thanos.TracingConfig,
+				},
+			})
 		}
-		if p.Spec.LogFormat != "" {
-			container.Args = append(container.Args, fmt.Sprintf("--log.format=%s", p.Spec.LogFormat))
+
+		if p.Spec.Thanos.LogLevel != "" {
+			container.Args = append(container.Args, "--log.level="+p.Spec.Thanos.LogLevel)
+		} else if p.Spec.LogLevel != "" {
+			container.Args = append(container.Args, "--log.level="+p.Spec.LogLevel)
+		}
+		if p.Spec.Thanos.LogFormat != "" {
+			container.Args = append(container.Args, "--log.format="+p.Spec.Thanos.LogFormat)
+		} else if p.Spec.LogFormat != "" {
+			container.Args = append(container.Args, "--log.format="+p.Spec.LogFormat)
+		}
+
+		if p.Spec.Thanos.MinTime != "" {
+			container.Args = append(container.Args, "--min-time="+p.Spec.Thanos.MinTime)
 		}
 		additionalContainers = append(additionalContainers, container)
 	}
@@ -840,33 +767,35 @@ func makeStatefulSetSpec(p monitoringv1.Prometheus, c *Config, ruleConfigMapName
 		promArgs = append(promArgs, "--storage.tsdb.max-block-duration=2h")
 	}
 
-	// Version is used by default.
-	// If the tag is specified, we use the tag to identify the container image.
-	// If the sha is specified, we use the sha to identify the container image,
-	// as it has even stronger immutable guarantees to identify the image.
-	prometheusImage := fmt.Sprintf("%s:%s", p.Spec.BaseImage, p.Spec.Version)
-	if p.Spec.Tag != "" {
-		prometheusImage = fmt.Sprintf("%s:%s", p.Spec.BaseImage, p.Spec.Tag)
+	configReloaderArgs := []string{
+		fmt.Sprintf("--config-file=%s", path.Join(confDir, configFilename)),
+		fmt.Sprintf("--config-envsubst-file=%s", path.Join(confOutDir, configEnvsubstFilename)),
 	}
-	if p.Spec.SHA != "" {
-		prometheusImage = fmt.Sprintf("%s@sha256:%s", p.Spec.BaseImage, p.Spec.SHA)
+	configReloaderVolumeMounts := []v1.VolumeMount{
+		{
+			Name:      "config",
+			MountPath: confDir,
+		},
+		{
+			Name:      "config-out",
+			MountPath: confOutDir,
+		},
 	}
-	if p.Spec.Image != nil && *p.Spec.Image != "" {
-		prometheusImage = *p.Spec.Image
-	}
-
-	prometheusConfigReloaderResources := v1.ResourceRequirements{Limits: v1.ResourceList{}}
-	if c.ConfigReloaderCPU != "0" {
-		prometheusConfigReloaderResources.Limits[v1.ResourceCPU] = resource.MustParse(c.ConfigReloaderCPU)
-	}
-	if c.ConfigReloaderMemory != "0" {
-		prometheusConfigReloaderResources.Limits[v1.ResourceMemory] = resource.MustParse(c.ConfigReloaderMemory)
+	if len(ruleConfigMapNames) != 0 {
+		for _, name := range ruleConfigMapNames {
+			mountPath := rulesDir + "/" + name
+			configReloaderVolumeMounts = append(configReloaderVolumeMounts, v1.VolumeMount{
+				Name:      name,
+				MountPath: mountPath,
+			})
+			configReloaderArgs = append(configReloaderArgs, fmt.Sprintf("--watched-dir=%s", mountPath))
+		}
 	}
 
 	operatorContainers := append([]v1.Container{
 		{
 			Name:                     "prometheus",
-			Image:                    prometheusImage,
+			Image:                    prometheusImagePath,
 			Ports:                    ports,
 			Args:                     promArgs,
 			VolumeMounts:             promVolumeMounts,
@@ -874,30 +803,28 @@ func makeStatefulSetSpec(p monitoringv1.Prometheus, c *Config, ruleConfigMapName
 			ReadinessProbe:           readinessProbe,
 			Resources:                p.Spec.Resources,
 			TerminationMessagePolicy: v1.TerminationMessageFallbackToLogsOnError,
-		}, {
-			Name:                     "prometheus-config-reloader",
-			Image:                    c.PrometheusConfigReloaderImage,
-			TerminationMessagePolicy: v1.TerminationMessageFallbackToLogsOnError,
-			Env: []v1.EnvVar{
-				{
-					Name: "POD_NAME",
-					ValueFrom: &v1.EnvVarSource{
-						FieldRef: &v1.ObjectFieldSelector{FieldPath: "metadata.name"},
-					},
-				},
-			},
-			Command:      []string{"/bin/prometheus-config-reloader"},
-			Args:         configReloadArgs,
-			VolumeMounts: configReloadVolumeMounts,
-			Resources:    prometheusConfigReloaderResources,
 		},
+		operator.CreateConfigReloader(
+			c.ReloaderConfig,
+			url.URL{
+				Scheme: "http",
+				Host:   c.LocalHost + ":9090",
+				Path:   path.Clean(webRoutePrefix + "/-/reload"),
+			},
+			p.Spec.ListenLocal,
+			c.LocalHost,
+			p.Spec.LogFormat,
+			p.Spec.LogLevel,
+			configReloaderArgs,
+			configReloaderVolumeMounts,
+		),
 	}, additionalContainers...)
 
 	containers, err := k8sutil.MergePatchContainers(operatorContainers, p.Spec.Containers)
 	if err != nil {
 		return nil, errors.Wrap(err, "failed to merge containers spec")
 	}
-	// PodManagementPolicy is set to Parallel to mitigate issues in kuberentes: https://github.com/kubernetes/kubernetes/issues/60164
+	// PodManagementPolicy is set to Parallel to mitigate issues in kubernetes: https://github.com/kubernetes/kubernetes/issues/60164
 	// This is also mentioned as one of limitations of StatefulSets: https://kubernetes.io/docs/concepts/workloads/controllers/statefulset/#limitations
 	return &appsv1.StatefulSetSpec{
 		ServiceName:         governingServiceName,
@@ -907,7 +834,7 @@ func makeStatefulSetSpec(p monitoringv1.Prometheus, c *Config, ruleConfigMapName
 			Type: appsv1.RollingUpdateStatefulSetStrategyType,
 		},
 		Selector: &metav1.LabelSelector{
-			MatchLabels: finalLabels,
+			MatchLabels: finalSelectorLabels,
 		},
 		Template: v1.PodTemplateSpec{
 			ObjectMeta: metav1.ObjectMeta{
@@ -917,7 +844,7 @@ func makeStatefulSetSpec(p monitoringv1.Prometheus, c *Config, ruleConfigMapName
 			Spec: v1.PodSpec{
 				Containers:                    containers,
 				InitContainers:                p.Spec.InitContainers,
-				SecurityContext:               securityContext,
+				SecurityContext:               p.Spec.SecurityContext,
 				ServiceAccountName:            p.Spec.ServiceAccountName,
 				NodeSelector:                  p.Spec.NodeSelector,
 				PriorityClassName:             p.Spec.PriorityClassName,
@@ -951,5 +878,9 @@ func subPathForStorage(s *monitoringv1.StorageSpec) string {
 		return ""
 	}
 
+	if s.DisableMountSubPath {
+		return ""
+	}
+
 	return "prometheus-db"
 }
